//
// Program for lecturing purposes by Marek Vinkler (xvinkl@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//

#include <memory>
#include <cmath>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>

// Include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

//----------------------------------------------------------------
// TODO:
// 1) Get location of the "position" attribute.
// 2) Create VAO containing the triangle geometry.
// 3) Use the programe and render the VAO.
// 4) Get location of the "m_pvm" uniform.
// 5) Set the "m_pvm" uniform with projection matrix.
// 6) Set fragment data location for varible "outputF".
// 7) Rotate the projection matrix.

//----------------------------------------------------------------

// Projection matrix, PVM = projection-view-model
float pvm_matrix[16];

// Shaders and programs
GLuint vertex_shader, fragment_shader, program;

// Vertex attribute
GLint VERTEX_COORD_ATTRIB;

// Uniforms
GLint UNIFORM_MODELVIEWPROJ;

// Vao
GLuint vao;

// Buffers
GLuint buffers[2];

// ------------------------------------------------------------
//
// Auxiliary functions
//

/// Performs matrix multiplication, computes C = A * B
void multMatrix(float *out_C, const float *in_A, const float *in_B)
{
	for (int col = 0; col < 4; ++col)
	{
		for (int row = 0; row < 4; ++row)
		{
			out_C[row*4 + col] = 0.0f;
			for (int k = 0; k < 4; ++k)
			{
				out_C[row*4 + col] += in_A[row*4 + k] * in_B[k*4 + col]; 
			}
		}
	}
}

/// Creates a projection matrix
void projectionMatrix(float *result, float fovy_degrees, float aspect, float near_plane, float far_plane)
{
	float fovy_radians = (fovy_degrees * (3.14f / 180.0f));
	float focus = 1.0f / tan (fovy_radians / 2.0f);
	float np = near_plane;
	float fp = far_plane;

	result[0] = focus/aspect;	result[1] = 0.0f;		result[2] = 0.0f;						result[3] = 0.0f;
	result[4] = 0.0f;			result[5] = focus;		result[6] = 0.0f;						result[7] = 0.0f;
	result[8] = 0.0f;			result[9] = 0.0f;		result[10] = -(fp+np)/(np-fp);			result[11] = -1.0f;
	result[12] = 0.0f;			result[13] = 0.0f;		result[14] = -(2.0f*fp*np)/(np-fp);		result[15] = 0.0f;
}

/// Creates a rotation matrix, rotating around Z axis
void rotateMatrix(float *result, float angle_degrees)
{
	float angle_radians = (angle_degrees * (3.14f / 180.0f));
	float s = sinf(angle_radians);
	float c = cosf(angle_radians);

	result[0] = c;		result[1] = s;		result[2] = 0.0f;		result[3] = 0.0f;
	result[4] = -s;		result[5] = c;		result[6] = 0.0f;		result[7] = 0.0f;
	result[8] = 0.0f;	result[9] = 0.0f;	result[10] = 1.0f;		result[11] = 0.0f;
	result[12] = 0.0f;	result[13] = 0.0f;	result[14] = 0.0f;		result[15] = 1.0f;
}

/// Loads a file from 'file_name' and returns it as a string
std::string load_file_to_string(const char *file_name)
{
	std::ifstream file(file_name);
	std::stringstream ss;
	ss << file.rdbuf();
	return ss.str();
}

/// Compiles given shader and checks for its compile status. If the status is OK,
/// it returns true. If there was a problem with the compilation, the log with
/// the information about the problem is obtained and printed, and the function
/// returns false.
///
/// 'shader_type' is the type of the shader, e.g. "vertex", "fragment", ...
bool compileAndCheckShader(GLuint shader_obj, const char *shader_type)
{
	glCompileShader(shader_obj);

	int compile_status;
	glGetShaderiv(shader_obj, GL_COMPILE_STATUS, &compile_status);
	if (!compile_status)
	{
		int log_len = 0;
		glGetShaderiv(shader_obj, GL_INFO_LOG_LENGTH, &log_len);
		std::unique_ptr<char []> log(new char[log_len]);
		glGetShaderInfoLog(shader_obj, log_len, nullptr, log.get());

		std::cout << "Failed to compile " << shader_type << " shader" << std::endl;
		std::cout << log.get() << std::endl;
		return false;
	}
	return true;
}

/// Links given program and checks for its link status. If the status is OK,
/// it returns true. If there was a problem with the linkage, the log with
/// the information about the problem is obtained and printed, and the function
/// returns false.
bool linkAndCheckProgram(GLuint program_obj)
{
	glLinkProgram(program_obj);
	
	int link_status;
	glGetProgramiv(program_obj, GL_LINK_STATUS, &link_status);
	if (!link_status)
	{
		int log_len = 0;
		glGetProgramiv(program_obj, GL_INFO_LOG_LENGTH, &log_len);
		std::unique_ptr<char []> log(new char[log_len]);
		glGetProgramInfoLog(program_obj, log_len, nullptr, log.get());

		std::cout << "Failed to link program" << std::endl;
		std::cout << log.get() << std::endl;
		return false;
	}
	return true;
}

// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;
	}

	// uncomment this if not using an idle func
	//glutPostRedisplay();
}

// ------------------------------------------------------------
//
// Render Scene
//

void renderScene(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// TODO: Use our program
	
	// TODO: Set uniforms
	
	// TODO: Render

	// Swap buffers
	glutSwapBuffers();
}

// --------------------------------------------------------
//
// Initialize OpenGL and other stuff
//

void setupShaders()
{
	vertex_shader = glCreateShader(GL_VERTEX_SHADER);
	fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);	

	std::string vs_source = load_file_to_string("shaders/helloWorld.vert");
	std::string fs_source = load_file_to_string("shaders/helloWorld.frag");

	const char * vv = vs_source.c_str();
	const char * ff = fs_source.c_str();

	glShaderSource(vertex_shader, 1, &vv, NULL);
	glShaderSource(fragment_shader, 1, &ff, NULL);

	compileAndCheckShader(vertex_shader, "vertex");
	compileAndCheckShader(fragment_shader, "fragment");

	program = glCreateProgram();

	glAttachShader(program, vertex_shader);
	glAttachShader(program, fragment_shader);

	// TODO: Set fragment data location

	linkAndCheckProgram(program);

	// TODO: Get locations of vertex attributes and uniforms
	// Vertex attributes
	
	// Uniforms
}

void setupGeometry()
{
	int verticeCount = 3;
	int indicesCount = 3;

	float vertices[] = {
		-1.0f, 0.0f, -2.0f, 1.0f,
		1.0f, 0.0f, -2.0f, 1.0f,
		0.0f, 1.0f, -2.0f, 1.0f
	};

	unsigned int indices[] = {
		0, 1, 2
	};

	// TODO: Setup vertex attributes
}

void setupCamera()
{
	// TODO: Define projection matrix
}

void initOpenGL()
{
	// Some GL settings
	glEnable(GL_DEPTH_TEST);
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
}


// ------------------------------------------------------------
//
// Main function
//

/// Callback function to be called when we make an error in OpenGL
void APIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
	printf(message);
}

int main(int argc, char **argv)
{
	// GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA);

	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(512,512);
	glutCreateWindow("PV227 - Triangle");

	// Callback Registration
	glutDisplayFunc(renderScene);
	glutIdleFunc(renderScene);

	// Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);

	// Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

	// Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));

	// Setup callback that will inform us when we make an error.
	// glDebugMessageCallbackARB is sometimes missed by glew, due to a bug in it.
	// Use (PFNGLDEBUGMESSAGECALLBACKARBPROC)glXGetProcAddress((unsigned char *)"glDebugMessageCallbackARB"); in Linux
	PFNGLDEBUGMESSAGECALLBACKARBPROC myglDebugMessageCallbackARB =
		(PFNGLDEBUGMESSAGECALLBACKARBPROC)wglGetProcAddress("glDebugMessageCallbackARB");
	if (myglDebugMessageCallbackARB)
	{
		glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
		myglDebugMessageCallbackARB(simple_debug_callback, nullptr);
	}

	initOpenGL();
	setupShaders();
	setupGeometry();
	setupCamera();

	// GLUT main loop
	glutMainLoop();

	// Clean up shaders
	glDetachShader(program, vertex_shader);
	glDetachShader(program, fragment_shader);
	glDeleteShader(vertex_shader);
	glDeleteShader(fragment_shader);
	glDeleteProgram(program);

	// Clean up geometry
	glDeleteVertexArrays(1, &vao);
	glDeleteBuffers(2, buffers);

	return 0;
}
